/*
 * omg-tcpdump
 *
 * Copyright (c) 2014 Johannes Noe
 * Licensed under the MIT license.
 */

'use strict';

var _ = require("lodash"),
    path = require("path"),
    spawn = require('child_process').spawn,
    environment = require("/data/config.json").environment,
    omg = {},
    modulesNeeded = ['logger'];

if (environment === "dev") {
    _.each(modulesNeeded, function(modName) {
        omg[modName] = require(path.normalize(__dirname + "/../../") + "omg-" + modName + "/");
    });
} else if (environment === "prod") {
    _.each(modulesNeeded, function(modName) {
        omg[modName] = require("omg-" + modName);
    });
}

var logger = omg.logger,
    tcpProcess,
    tcpTimer;

//kill tcpdump
exports.stop = function(startCallback, outputCallback) {
    if (tcpProcess !== undefined && tcpProcess.killed !== true) {
        tcpProcess.on("error", function(err) {
            logger.error(err.message);
        });
        tcpProcess.kill();
        if (outputCallback) {
            outputCallback(null, 'tcpdump killed');
        }
    } else {
        logger.error("tcpProcess is not defined, or already killed");
        if (outputCallback) {
            outputCallback(null, 'tcpdump killed');
        }
    }
    //clear timer that kills processes after some time
    if (tcpTimer) {
        logger.debug('clearing tcpdump timeout');
        clearTimeout(tcpTimer);
    }
};

//start tcpdump to /tmp/logs/capture.cap
exports.start = function(device, duration, startCallback, outputCallback) {
    if (arguments.length < 4) {
        logger.warn("You did not set all required arguments! Trying to find out what's missing");
        var foundDevice, foundDuration, foundStartCallback, foundOutputCallback;
        _.forEach(arguments, function(arg) {
            var theType = typeof(arg);
            var reg = /^\d+$/;

            if (theType === "function") {
                if (foundStartCallback) {
                    outputCallback = arg;
                    logger.warn("found another function, assuming it's the outputCallback");
                } else {
                    startCallback = arg;
                    foundStartCallback = true;
                    logger.warn("found a function, assuming it's the startCallback");
                }
            }
            if (theType === "number" || reg.test(arg)) {
                duration = arg;
                foundDuration = true;
                logger.warn("found a number, assuming it's the duration: " + arg);
            }
            if (theType === "string" && !reg.test(arg)) {
                device = arg;
                logger.warn("found a string, assuming it's the device: " + arg);
                foundDevice = true;
            }
        });

        if (!foundStartCallback) {
            logger.error("did not find a callback, can't handle request");
            return;
        }
        if (!foundDevice) {
            device = 'eth0';
            logger.warn("did not find a device name, setting it to default: eth0");
        }
        if (!foundDuration) {
            duration = 15;
            logger.warn("did not find a duration value, setting it to default: 15 minutes");
        }
    }

    // if tcpdump is running, return
    if (tcpProcess && !tcpProcess.killed) {
        return startCallback("already running");
    }

    var command, filter, args;

    //get mac address, then start tcpdump
    require('getmac').getMac(function(err, macAddress) {
        if (err) {
            logger.error(err.message);
            return;
        }

        if (device.indexOf("eth") > -1) {
            filter = " ether host " + macAddress + " and not port 514 and not port 22";
        } else {
            device = "lo";
            filter = " not port 514 and not port 22 ";
        }

        command = "tcpdump";
        // args = ["-a","-U","-i" + device, "-w /tmp/logs/capture.cap", "-p", "-n", "-s 0", filter,  " < /dev/null >/dev/null 2>/dev/null"];
        args = ["-a", "-U", "-i", device, "-w", "/tmp/logs/capture.cap", "-p", "-n", "-s", "0", filter];

        logger.debug("running " + command + ' ' + args.join(' '));

        tcpProcess = spawn(command, args);

        tcpProcess.stdout.on('data', function(data) {
            logger.debug('stdout: ' + data);
        });
        tcpProcess.stderr.on('data', function(data) {
            logger.debug('sterr: ' + data);
            if (data.toString().indexOf("listening on") > -1) {
                outputCallback(null, data);

                //reset tcpTimer
                if (tcpTimer) {
                    logger.debug("clearing tcpdump timeout");
                    clearTimeout(tcpTimer);
                }

                //set tcpTimer
                logger.debug("setting tcpdump timeout to " + duration + " minutes");

                tcpTimer = setTimeout(function() {
                    exports.stop(outputCallback);
                }, duration * 1000 * 60);
            }
        });
    });
};